#ifndef TYPON_CONTINUATION_HPP_INCLUDED
#define TYPON_CONTINUATION_HPP_INCLUDED

#include <atomic>
#include <coroutine>
#include <cstdint>
#include <exception>
#include <limits>


namespace typon
{

  struct coroutine_node
  {
    coroutine_node * _next;
    std::coroutine_handle<> _coroutine;
    std::exception_ptr * _exception;
  };


  struct coroutine_list
  {
    coroutine_node * _first = nullptr;

    template <typename Promise>
    void insert(std::coroutine_handle<Promise> coroutine) noexcept
    {
      std::exception_ptr * exception = &(coroutine.promise()._exception);
      coroutine.promise()._node = { _first, coroutine, exception };
      _first = &(coroutine.promise()._node);
    }

    void check_exceptions()
    {
      for (auto node = _first; node != nullptr; node = node->_next)
      {
        std::exception_ptr & exception = *(node->_exception);
        if (exception)
        {
          std::rethrow_exception(exception);
        }
      }
    }

    void clear() noexcept
    {
      auto next = _first;
      while(next)
      {
        auto node = next;
        next = next->_next;
        node->_coroutine.destroy();
      }
      _first = nullptr;
    }
  };


  struct continuation_data
  {
    using u64 = std::uint_fast64_t;
    static constexpr u64 UMAX = std::numeric_limits<u64>::max();

    std::coroutine_handle<> _coroutine;
    std::coroutine_handle<> _continuation;
    coroutine_list _children;

    std::atomic<u64> _n = UMAX;
    u64 _thefts = 0;

    continuation_data(std::coroutine_handle<> coroutine)
      : _coroutine(coroutine)
    {}
  };


  struct continuation_handle
  {
    continuation_data * _data;

    continuation_handle() noexcept {}

    template <typename Promise>
    continuation_handle(std::coroutine_handle<Promise> coroutine) noexcept
      : _data(&(coroutine.promise()._data))
    {}

    auto & thefts() noexcept
    {
      return _data->_thefts;
    }

    auto & children() noexcept
    {
      return _data->_children;
    }

    auto & n() noexcept
    {
      return _data->_n;
    }

    void resume()
    {
      _data->_coroutine.resume();
    }

    operator std::coroutine_handle<>() noexcept
    {
      return _data->_coroutine;
    }

    std::coroutine_handle<> continuation() noexcept
    {
      if (_data->_coroutine.done())
      {
        return _data->_continuation;
      }
      return _data->_coroutine;
    }
  };

}

#endif // TYPON_CONTINUATION_HPP_INCLUDED
