#ifndef TYPON_DEQUE_HPP_INCLUDED
#define TYPON_DEQUE_HPP_INCLUDED

#include <atomic>
#include <cstdint>
#include <memory>
#include <utility>


namespace typon
{

  template <typename T>
  struct RingBuffer
  {
    using u8 = std::uint_least8_t;
    using u64 = std::uint_fast64_t;

    using enum std::memory_order;

    const u8 _bits;
    const u64 _mask;
    RingBuffer * const _next;
    std::atomic<T> * const _array;

    RingBuffer(u8 bits, RingBuffer * next = nullptr) noexcept
      : _bits(bits)
      , _mask(this->capacity() - 1)
      , _next(next)
      , _array(new std::atomic<T>[this->capacity()])
    {}

    ~RingBuffer()
    {
      delete [] _array;
      if (_next)
      {
        delete _next;
      }
    }

    u64 capacity() noexcept
    {
      return u64(1) << _bits;
    }

    void put(u64 index, T object) noexcept
    {
      _array[index & _mask].store(std::move(object), relaxed);
    }

    T get(u64 index) noexcept
    {
      return _array[index & _mask].load(relaxed);
    }

    RingBuffer * fill(RingBuffer * sink, u64 start, u64 end) noexcept
    {
      for (u64 i = start; i < end; i++)
      {
        sink->put(i, get(i));
      }
      return sink;
    }

    RingBuffer * grow(u64 start, u64 end) noexcept
    {
      return fill(new RingBuffer(_bits + 1, this), start, end);
    }

    RingBuffer * shrink(u64 start, u64 end) noexcept
    {
      return fill(std::exchange(_next, nullptr), start, end);
    }
  };


  template <typename T>
  struct Optional
  {
    enum State : unsigned char { empty = 0, abort = 1, engaged = 2};

    State _state;
    union
    {
      T _value;
    };

    Optional(State state = empty) noexcept : _state(state) {}

    Optional(T value) noexcept : _state(engaged), _value(value) {}

    ~Optional()
    {
      if (_state == engaged)
      {
        std::destroy_at(std::addressof(_value));
      }
    }

    operator bool() noexcept
    {
      return _state == engaged;
    }

    State state() noexcept
    {
      return _state;
    }

    T * operator->() noexcept
    {
      return std::addressof(_value);
    }

    T & operator*() noexcept
    {
      return _value;
    }
  };


  template <typename T>
  struct Deque
  {
    using u8 = typename RingBuffer<T>::u8;
    using u64 = typename RingBuffer<T>::u64;
    using Array = RingBuffer<T>;

    using enum std::memory_order;

    std::atomic<u64> _top {1};
    std::atomic<u64> _bottom {1};
    std::atomic<Array *> _array;

    Deque(u8 bits = 2) noexcept
      : _array(new Array(bits))
    {}

    ~Deque()
    {
      delete _array;
    }

    void push(T x) noexcept
    {
      u64 t = _top.load(relaxed);
      u64 b = _bottom.load(acquire);
      Array * array = _array.load(relaxed);
      if (t - b > array->capacity() - 1)
      {
        array = array->grow(b, t);
        _array.store(array);
      }
      array->put(t, x);
      std::atomic_thread_fence(release);
      _top.store(t + 1, relaxed);
    }

    Optional<T> pop() noexcept
    {
      u64 t = _top.load(relaxed) - 1;
      Array * array = _array.load(relaxed);
      _top.store(t, relaxed);
      std::atomic_thread_fence(seq_cst);
      u64 b = _bottom.load(relaxed);
      Optional<T> x {};
      if (b <= t)
      {
        x = array->get(t);
        if (b == t)
        {
          if (!_bottom.compare_exchange_strong(b, b + 1, seq_cst, relaxed))
          {
            x = {};
          }
          _top.store(t + 1, relaxed);
        }
      }
      else
      {
        _top.store(t + 1, relaxed);
      }
      return x;
    }

    Optional<T> steal() noexcept
    {
      u64 b = _bottom.load(acquire);
      std::atomic_thread_fence(seq_cst);
      u64 t = _top.load(acquire);
      Optional<T> x {};
      if (b < t)
      {
        Array * array = _array.load(consume);
        x = array->get(b);
        if (!_bottom.compare_exchange_strong(b, b + 1, seq_cst, relaxed))
        {
          return {Optional<T>::abort};
        }
      }
      return x;
    }
  };

}


#endif // TYPON_DEQUE_HPP_INCLUDED
