#ifndef TYPON_FUTURE_HPP_INCLUDED
#define TYPON_FUTURE_HPP_INCLUDED

#include <coroutine>
#include <memory>
#include <type_traits>
#include <utility>

#include <typon/fundamental/meta.hpp>
#include <typon/fundamental/scope.hpp>

#include <typon/result.hpp>


namespace typon
{

  template <typename Promise>
  struct Future
  {
    using value_type = typename Promise::value_type;

    std::coroutine_handle<Promise> _coroutine;
    union
    {
      value_type _result;
    };

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        fdt::defer defer { [&coroutine]() { coroutine.destroy(); } };
        std::construct_at(&(_result), coroutine.promise().get());
      }
      else
      {
        _coroutine = coroutine;
      }
    }

    Future(Future && other)
      noexcept(std::is_nothrow_move_constructible_v<value_type>)
    {
      _coroutine = other._coroutine;
      if (!_coroutine)
      {
        std::construct_at(std::addressof(_result), std::move(other._result));
      }
    }

    Future& operator=(Future && other)
      noexcept(std::is_nothrow_move_constructible_v<value_type>)
    {
      if (this != &other)
      {
        Future old { std::move(*this) };
        _coroutine = other._coroutine;
        if (!_coroutine)
        {
          std::construct_at(std::addressof(_result), std::move(other._result));
        }
      }
      return *this;
    }

    ~Future()
    {
      if (!_coroutine)
      {
        std::destroy_at(&(_result));
      }
    }

    auto get()
    {
      if (_coroutine)
      {
        return _coroutine.promise().get();
      }
      return _result;
    }
  };


  template <typename Promise>
    requires requires { sizeof(typename Promise::value_type); }
    && (sizeof(typename Promise::value_type) > 2 * sizeof(void*))
  struct Future<Promise>
  {
    std::coroutine_handle<Promise> _coroutine;
    bool _owning;

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      _coroutine = coroutine;
      _owning = ready;
      if (ready)
      {
        if (coroutine.promise()._exception)
        {
          std::rethrow_exception(coroutine.promise()._exception);
        }
      }
    }

    Future(const Future &) = delete;
    Future& operator=(const Future &) = delete;

    Future(Future && other) noexcept
      : _coroutine(other._coroutine)
      , _owning(std::exchange(other._owning, false))
    {}

    Future& operator=(Future && other) noexcept
    {
      std::swap(_owning, other._owning);
      std::swap(_coroutine, other._coroutine);
      return *this;
    }

    ~Future()
    {
      if (_owning)
      {
        _coroutine.destroy();
      }
    }

    auto get()
    {
      return _coroutine.promise().get();
    }
  };


  template <typename Promise>
    requires std::is_trivially_copyable_v<typename Promise::value_type>
  struct Future<Promise>
  {
    using value_type = typename Promise::value_type;

    std::coroutine_handle<Promise> _coroutine;
    union
    {
      value_type _result;
    };

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        fdt::defer defer { [&coroutine]() { coroutine.destroy(); } };
        std::construct_at(&(_result), coroutine.promise().get());
      }
      else
      {
        _coroutine = coroutine;
      }
    }

    Future(Future && other) = default;
    Future& operator=(Future && other) = default;

    ~Future()
    {
      if (!_coroutine)
      {
        std::destroy_at(&(_result));
      }
    }

    auto get()
    {
      if (_coroutine)
      {
        return _coroutine.promise().get();
      }
      return _result;
    }
  };


  template <typename Promise>
    requires std::is_reference_v<typename Promise::value_type>
  struct Future<Promise>
  {
    using value_type = typename Promise::value_type;

    std::coroutine_handle<Promise> _coroutine;
    std::remove_reference<value_type> * _result;

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        fdt::defer defer { [&coroutine]() { coroutine.destroy(); } };
        _result = std::addressof(coroutine.promise().get());
      }
      else
      {
        _coroutine = coroutine;
      }
    }

    value_type get()
    {
      if (_coroutine)
      {
        return _coroutine.promise().get();
      }
      return *_result;
    }
  };


  template <typename Promise>
    requires std::is_void_v<typename Promise::value_type>
  struct Future<Promise>
  {
    using value_type = typename Promise::value_type;

    std::coroutine_handle<Promise> _coroutine;

    Future(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        fdt::defer defer { [&coroutine]() { coroutine.destroy(); } };
        coroutine.promise().get();
      }
      else
      {
        _coroutine = coroutine;
      }
    }

    void get()
    {
      if (_coroutine)
      {
        _coroutine.promise().get();
      }
    }
  };

}


#endif // TYPON_FUTURE_HPP_INCLUDED
