#ifndef TYPON_SCHEDULER_HPP_INCLUDED
#define TYPON_SCHEDULER_HPP_INCLUDED

#include <atomic>
#include <cstdint>
#include <exception>
#include <thread>
#include <vector>

#include <typon/continuation.hpp>
#include <typon/eventcount.hpp>
#include <typon/deque.hpp>
#include <typon/random.hpp>


namespace typon
{

  struct Scheduler
  {
    using uint = unsigned int;
    using Task = Optional<continuation_handle>;
    using Deque = Deque<continuation_handle>;

    static inline thread_local uint thread_id;

    std::atomic<uint> _actives = 0;
    std::atomic<uint> _thieves = 0;
    std::vector<Deque> _deque;
    std::vector<std::thread> _thread;
    std::atomic_bool _done {false};
    EventCount<> _notifyer;
    const uint _parallelism;

    static Scheduler & get() noexcept
    {
      static Scheduler scheduler {std::thread::hardware_concurrency()};
      return scheduler;
    }

    static void push(continuation_handle task) noexcept
    {
      get()._deque[thread_id].push(task);
    }

    static void schedule(continuation_handle task) noexcept
    {
      Scheduler & scheduler = get(); 
      scheduler._deque[thread_id].push(task);
      scheduler._notifyer.notify_one();
    }

    static Task pop() noexcept
    {
      return get()._deque[thread_id].pop();
    }

    Scheduler(uint parallelism) noexcept
      : _deque(parallelism + 1)
      , _parallelism(parallelism)
    {
      thread_id = parallelism;

      for (uint id = 0; id < parallelism; id++)
      {
        _thread.emplace_back([this, id]() {
          thread_id = id;

          Task task {};
          for(;;)
          {
            exploit_task(task);
            if (!wait_for_task(task))
            {
              break;
            }
          }
        });
      }
    }

    ~Scheduler() noexcept
    {
      _done.store(true);
      _notifyer.notify_all();
      for (auto & t : _thread)
      {
        t.join();
      }
    }

    void exploit_task(Task & task) noexcept
    {
      if (task)
      {
        if (_actives.fetch_add(1) == 0)
        {
          if (_thieves.load() == 0)
          {
            _notifyer.notify_one();
          }
        }
        while(task)
        {
          task->resume();
          task = _deque[thread_id].pop();
          if (task)
          {
            printf("[%u] ERROR unexpectedly pop valid task %p\n", thread_id, task->_data);
          }
        }
        _actives.fetch_sub(1);
      }
    }

    void explore_task(Task & task) noexcept
    {
      for (uint i = 0; i < _parallelism * 2 + 1; i++)
      {
        uint id = random::random() % _parallelism;
        if (id == thread_id)
        {
          task = _deque.back().steal();
        }
        else
        {
          task = _deque[id].steal();
        }
        if (task)
        {
          task->thefts()++;
          break;
        }
      }
    }

    bool wait_for_task(Task & task) noexcept
    {
    wait_for_task:
      _thieves.fetch_add(1);
    explore_task:
      explore_task(task);
      if (task)
      {
        if (_thieves.fetch_sub(1) == 1)
        {
          _notifyer.notify_one();
        }
        return true;
      }
      auto key = _notifyer.prepare_wait();
      task = _deque.back().steal();
      if (task.state() != Task::empty)
      {
        _notifyer.cancel_wait();
        if (task)
        {
          if (_thieves.fetch_sub(1) == 1)
          {
            _notifyer.notify_one();
          }
          return true;
        }
        goto explore_task;
      }
      if (_done.load())
      {
        _notifyer.cancel_wait();
        _notifyer.notify_all();
        _thieves.fetch_sub(1);
        return false;
      }
      if (_thieves.fetch_sub(1) == 1)
      {
        if (_actives.load() > 0)
        {
          _notifyer.cancel_wait();
          goto wait_for_task;
        }
      }
      _notifyer.wait(key);
      return true;
    }
  };

}

#endif // TYPON_SCHEDULER_HPP_INCLUDED
